# Copyright 2022, Christian Pehle
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as np
import tree_math

from jaxsnn.base.types import ArrayLike
import dataclasses


@dataclasses.dataclass
@tree_math.struct
class HodgekinHuxleyState:
    v: ArrayLike
    m: ArrayLike
    h: ArrayLike
    n: ArrayLike


@tree_math.struct
class HodgekinHuxleyParameters:
    g_K: ArrayLike
    E_K: ArrayLike
    g_Na: ArrayLike
    E_Na: ArrayLike
    g_L: ArrayLike
    E_L: ArrayLike
    C_m_inv: ArrayLike


def alpha_n(Vm):
    return (0.02 * (Vm - 25.0)) / (1.0 - np.exp((-1.0 * (Vm - 25.0)) / 9.0))


def beta_n(Vm):
    return (-0.002 * (Vm - 25.0)) / (1.0 - np.exp((Vm - 25.0) / 9.0))


def alpha_m(Vm):
    return (0.182 * (Vm + 35.0)) / (1.0 - np.exp((-1.0 * (Vm + 35.0)) / 9.0))


def beta_m(Vm):
    return (-0.124 * (Vm + 35.0)) / (1.0 - np.exp((Vm + 35.0) / 9.0))


def alpha_h(Vm):
    return 0.25 * np.exp((-1.0 * (Vm + 90.0)) / 12.0)


def beta_h(Vm):
    return (0.25 * np.exp((Vm + 62.0) / 6.0)) / np.exp((Vm + 90.0) / 12.0)


def I_Na(s: HodgekinHuxleyState, p: HodgekinHuxleyParameters):
    return p.g_Na * np.power(s.m, 3) * s.h * (s.v - p.E_Na)


def I_K(s: HodgekinHuxleyState, p: HodgekinHuxleyParameters):
    return p.g_K * np.power(s.n, 4) * (s.v - p.E_K)


def I_L(s: HodgekinHuxleyState, p: HodgekinHuxleyParameters):
    return p.g_L * (s.v - p.E_L)


def channel_dynamics(alpha, beta):
    def dynamics(x: ArrayLike, v: ArrayLike):
        return alpha(v) * (1 - x) - beta(v) * x

    return dynamics


def hodgekin_huxley_dynamics(p: HodgekinHuxleyParameters):
    m_dynamics = channel_dynamics(alpha_m, beta_m)
    h_dynamics = channel_dynamics(alpha_h, beta_h)
    n_dynamics = channel_dynamics(alpha_n, beta_n)

    def dynamics(s: HodgekinHuxleyState, I: ArrayLike):
        return HodgekinHuxleyState(
            v=-p.C_m_inv * (I_K(s, p) + I_Na(s, p) + I_L(s, p) - I),
            m=m_dynamics(s.m, s.v),
            n=n_dynamics(s.n, s.v),
            h=h_dynamics(s.h, s.v),
        )

    return dynamics



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for implicit time_stepping."""

import jax
import jax.numpy as jnp
import numpy as np
from absl.testing import absltest, parameterized
from jax import tree_util
from njx.base import funcutils
from njx.base import implicit as time_stepping

jax.config.update("jax_enable_x64", True)


def harmonic_oscillator(x0, t):
    theta = jnp.arctan(x0[0] / x0[1])
    r = jnp.linalg.norm(x0, ord=2, axis=0)
    return r * jnp.stack([jnp.sin(t + theta), jnp.cos(t + theta)])


class CustomODE(time_stepping.ImplicitExplicitODE):
    def __init__(self, explicit_terms, implicit_terms, implicit_solve):
        self.explicit_terms = explicit_terms
        self.implicit_terms = implicit_terms
        self.implicit_solve = implicit_solve


ALL_TEST_PROBLEMS = [
    # x(t) = np.ones(10)
    dict(
        testcase_name="_zero_derivative",
        explicit_terms=lambda x: 0 * x,
        implicit_terms=lambda x: 0 * x,
        implicit_solve=lambda x, eta: x,
        dt=1e-2,
        inner_steps=10,
        outer_steps=5,
        initial_state=np.ones(10),
        closed_form=lambda x0, t: x0,
        tolerances=[1e-12] * 5,
    ),
    # x(t) = 5 * t * np.ones(3)
    dict(
        testcase_name="_constant_derivative",
        explicit_terms=lambda x: 5 * jnp.ones_like(x),
        implicit_terms=lambda x: 0 * x,
        implicit_solve=lambda x, eta: x,
        dt=1e-2,
        inner_steps=10,
        outer_steps=5,
        initial_state=np.ones(3),
        closed_form=lambda x0, t: x0 + 5 * t,
        tolerances=[1e-12] * 5,
    ),
    # x(t) = np.arange(3) * np.exp(t)
    # Uses explicit terms only.
    dict(
        testcase_name="_linear_derivative_explicit",
        explicit_terms=lambda x: x,
        implicit_terms=lambda x: 0 * x,
        implicit_solve=lambda x, eta: x,
        dt=1e-2,
        inner_steps=20,
        outer_steps=5,
        initial_state=np.arange(3.0),
        closed_form=lambda x0, t: np.arange(3) * jnp.exp(t),
        tolerances=[5e-2, 1e-4, 1e-6, 1e-9, 1e-6],
    ),
    # x(t) = np.arange(3) * np.exp(t)
    # Uses implicit terms only.
    dict(
        testcase_name="_linear_derivative_implicit",
        explicit_terms=lambda x: 0 * x,
        implicit_terms=lambda x: x,
        implicit_solve=lambda x, eta: x / (1 - eta),
        dt=1e-2,
        inner_steps=20,
        outer_steps=5,
        initial_state=np.arange(3.0),
        closed_form=lambda x0, t: np.arange(3) * jnp.exp(t),
        tolerances=[5e-2, 5e-5, 1e-5, 1e-5, 3e-5],
    ),
    # x(t) = np.arange(3) * np.exp(t)
    # Splits the equation into an implicit and explicit term.
    dict(
        testcase_name="_linear_derivative_semi_implicit",
        explicit_terms=lambda x: x / 2,
        implicit_terms=lambda x: x / 2,
        implicit_solve=lambda x, eta: x / (1 - eta / 2),
        dt=1e-2,
        inner_steps=20,
        outer_steps=5,
        initial_state=np.arange(3) * np.exp(0),
        closed_form=lambda x0, t: np.arange(3.0) * jnp.exp(t),
        tolerances=[1e-4, 2e-5, 2e-6, 1e-6, 2e-5],
    ),
    dict(
        testcase_name="_harmonic_oscillator_explicit",
        explicit_terms=lambda x: jnp.stack([x[1], -x[0]]),
        implicit_terms=jnp.zeros_like,
        implicit_solve=lambda x, eta: x,
        dt=1e-2,
        inner_steps=20,
        outer_steps=5,
        initial_state=np.ones(2),
        closed_form=harmonic_oscillator,
        tolerances=[1e-2, 3e-5, 6e-8, 5e-11, 6e-8],
    ),
    dict(
        testcase_name="_harmonic_oscillator_implicit",
        explicit_terms=jnp.zeros_like,
        implicit_terms=lambda x: jnp.stack([x[1], -x[0]]),
        implicit_solve=lambda x, eta: jnp.stack(  # pylint: disable=g-long-lambda
            [x[0] + eta * x[1], x[1] - eta * x[0]]
        )
        / (1 + eta**2),
        dt=1e-2,
        inner_steps=20,
        outer_steps=5,
        initial_state=np.ones(2),
        closed_form=harmonic_oscillator,
        tolerances=[1e-2, 2e-5, 2e-6, 1e-6, 6e-6],
    ),
]


ALL_TIME_STEPPERS = [
    time_stepping.backward_forward_euler,
    time_stepping.crank_nicolson_rk2,
    time_stepping.crank_nicolson_rk3,
    time_stepping.crank_nicolson_rk4,
    time_stepping.imex_rk_sil3,
]


class TimeSteppingTest(parameterized.TestCase):
    @parameterized.named_parameters(ALL_TEST_PROBLEMS)
    def test_implicit_solve(
        self,
        explicit_terms,
        implicit_terms,
        implicit_solve,
        dt,
        inner_steps,
        outer_steps,
        initial_state,
        closed_form,
        tolerances,
    ):
        """Tests that time integration is accurate for a range of test cases."""
        del dt, explicit_terms, inner_steps, outer_steps, closed_form  # unused
        del tolerances  # unused

        # Verifies that `implicit_solve` solves (y - eta * F(y)) = x
        # This does not test the integrator, but rather verifies that the test
        # case is valid.
        eta = 0.3
        solved_state = implicit_solve(initial_state, eta)
        reconstructed_state = solved_state - eta * implicit_terms(solved_state)
        np.testing.assert_allclose(reconstructed_state, initial_state)

    @parameterized.named_parameters(ALL_TEST_PROBLEMS)
    def test_integration(
        self,
        explicit_terms,
        implicit_terms,
        implicit_solve,
        dt,
        inner_steps,
        outer_steps,
        initial_state,
        closed_form,
        tolerances,
    ):
        # Compute closed-form solution.
        time = dt * inner_steps * (1 + np.arange(outer_steps))
        expected = jax.vmap(closed_form, in_axes=(None, 0))(initial_state, time)

        # Compute trajectory using time-stepper.
        for atol, time_stepper in zip(tolerances, ALL_TIME_STEPPERS):
            with self.subTest(time_stepper.__name__):
                equation = CustomODE(explicit_terms, implicit_terms, implicit_solve)
                semi_implicit_step = time_stepper(equation, dt)
                integrator = funcutils.trajectory(
                    funcutils.repeated(semi_implicit_step, inner_steps), outer_steps
                )
                _, actual = integrator(initial_state)
                np.testing.assert_allclose(expected, actual, atol=atol, rtol=0)

    def test_pytree_state(self):
        equation = CustomODE(
            explicit_terms=lambda x: tree_util.tree_map(jnp.zeros_like, x),
            implicit_terms=lambda x: tree_util.tree_map(jnp.zeros_like, x),
            implicit_solve=lambda x, eta: x,
        )
        u0 = {"x": 1.0, "y": 1.0}
        for time_stepper in ALL_TIME_STEPPERS:
            with self.subTest(time_stepper.__name__):
                u1 = time_stepper(equation, 1.0)(u0)
                self.assertEqual(u0, u1)


if __name__ == "__main__":
    config.update("jax_enable_x64", True)
    absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implicit-explicit time stepping routines for ODEs."""
import dataclasses
from typing import Callable, Sequence, TypeVar
import tree_math


PyTreeState = TypeVar("PyTreeState")
ControlInput = TypeVar("ControlInput")

TimeStepFn = Callable[[PyTreeState], PyTreeState]
ControlledTimeStepFn = Callable[[PyTreeState, ControlInput], PyTreeState]


class ImplicitExplicitODE:
    """Describes a set of ODEs with implicit & explicit terms.
    The equation is given by:
      ∂x/∂t = explicit_terms(x) + implicit_terms(x)
    `explicit_terms(x)` includes terms that should use explicit time-stepping and
    `implicit_terms(x)` includes terms that should be modeled implicitly.
    Typically the explicit terms are non-linear and the implicit terms are linear.
    This simplifies solves but isn't strictly necessary.
    """

    def explicit_terms(self, state: PyTreeState) -> PyTreeState:
        """Evaluates explicit terms in the ODE."""
        raise NotImplementedError

    def implicit_terms(self, state: PyTreeState) -> PyTreeState:
        """Evaluates implicit terms in the ODE."""
        raise NotImplementedError

    def implicit_solve(
        self,
        state: PyTreeState,
        step_size: float,
    ) -> PyTreeState:
        """Solves `y - step_size * implicit_terms(y) = x` for y."""
        raise NotImplementedError


class ImplicitExplicitCDE:
    """Describes a set of CDEs with implicit & explicit terms.

    We assume that only the explicit terms are subject to control input.

    The equation is given by:
      ∂x/∂t = explicit_terms(x, u) + implicit_terms(x)
    `explicit_terms(x, u)` includes terms that should use explicit time-stepping and are controlled
    `implicit_terms(x)` includes terms that should be modeled implicitly.
    Typically the explicit terms are non-linear and the implicit terms are linear.
    This simplifies solves but isn't strictly necessary.
    """

    def explicit_terms(self, state: PyTreeState, u: ControlInput) -> PyTreeState:
        """Evaluates explicit terms in the ODE."""
        raise NotImplementedError

    def implicit_terms(self, state: PyTreeState) -> PyTreeState:
        """Evaluates implicit terms in the ODE."""
        raise NotImplementedError

    def implicit_solve(
        self,
        state: PyTreeState,
        step_size: float,
    ) -> PyTreeState:
        """Solves `y - step_size * implicit_terms(y) = x` for y."""
        raise NotImplementedError


def backward_forward_euler(
    equation: ImplicitExplicitODE,
    time_step: float,
) -> TimeStepFn:
    """Time stepping via forward and backward Euler methods.
    This method is first order accurate.
    Args:
      equation: equation to solve.
      time_step: time step.
    Returns:
      Function that performs a time step.
    """
    # pylint: disable=invalid-name
    dt = time_step
    F = tree_math.unwrap(equation.explicit_terms)
    G_inv = tree_math.unwrap(equation.implicit_solve, vector_argnums=0)

    @tree_math.wrap
    def step_fn(u0):
        g = u0 + dt * F(u0)
        u1 = G_inv(g, dt)
        return u1

    return step_fn


def backward_forward_euler_cde(
    equation: ImplicitExplicitCDE,
    time_step: float,
) -> TimeStepFn:
    """Time stepping via forward and backward Euler methods.
    This method is first order accurate.
    Args:
      equation: equation to solve.
      time_step: time step.
    Returns:
      Function that performs a time step.
    """
    # pylint: disable=invalid-name
    dt = time_step
    F = tree_math.unwrap(equation.explicit_terms)
    G_inv = tree_math.unwrap(equation.implicit_solve, vector_argnums=0)

    @tree_math.wrap
    def step_fn(u0, x0):
        g = u0 + dt * F(u0, x0)
        u1 = G_inv(g, dt)
        return u1, u1

    return step_fn


def crank_nicolson_rk2(
    equation: ImplicitExplicitODE,
    time_step: float,
) -> TimeStepFn:
    """Time stepping via Crank-Nicolson and 2nd order Runge-Kutta (Heun).
    This method is second order accurate.
    Args:
      equation: equation to solve.
      time_step: time step.
    Returns:
      Function that performs a time step.
    Reference:
      Chandler, G. J. & Kerswell, R. R. Invariant recurrent solutions embedded in
      a turbulent two-dimensional Kolmogorov flow. J. Fluid Mech. 722, 554–595
      (2013). https://doi.org/10.1017/jfm.2013.122 (Section 3)
    """
    # pylint: disable=invalid-name
    dt = time_step
    F = tree_math.unwrap(equation.explicit_terms)
    G = tree_math.unwrap(equation.implicit_terms)
    G_inv = tree_math.unwrap(equation.implicit_solve, vector_argnums=0)

    @tree_math.wrap
    def step_fn(u0):
        g = u0 + 0.5 * dt * G(u0)
        h1 = F(u0)
        u1 = G_inv(g + dt * h1, 0.5 * dt)
        h2 = 0.5 * (F(u1) + h1)
        u2 = G_inv(g + dt * h2, 0.5 * dt)
        return u2

    return step_fn


def crank_nicolson_rk2_cde(
    equation: ImplicitExplicitCDE,
    time_step: float,
) -> ControlledTimeStepFn:
    """Time stepping via Crank-Nicolson and 2nd order Runge-Kutta (Heun).
    This method is second order accurate.
    Args:
      equation: equation to solve.
      time_step: time step.
    Returns:
      Function that performs a time step.
    Reference:
      Chandler, G. J. & Kerswell, R. R. Invariant recurrent solutions embedded in
      a turbulent two-dimensional Kolmogorov flow. J. Fluid Mech. 722, 554–595
      (2013). https://doi.org/10.1017/jfm.2013.122 (Section 3)
    """
    # pylint: disable=invalid-name
    dt = time_step
    F = tree_math.unwrap(equation.explicit_terms)
    G = tree_math.unwrap(equation.implicit_terms)
    G_inv = tree_math.unwrap(equation.implicit_solve, vector_argnums=0)

    @tree_math.wrap
    def step_fn(u0, x0):
        g = u0 + 0.5 * dt * G(u0)
        h1 = F(u0, x0)
        u1 = G_inv(g + dt * h1, 0.5 * dt)
        h2 = 0.5 * (F(u1, x0) + h1)
        u2 = G_inv(g + dt * h2, 0.5 * dt)
        return u2, u2

    return step_fn


def low_storage_runge_kutta_crank_nicolson(
    alphas: Sequence[float],
    betas: Sequence[float],
    gammas: Sequence[float],
    equation: ImplicitExplicitODE,
    time_step: float,
) -> TimeStepFn:
    """Time stepping via "low-storage" Runge-Kutta and Crank-Nicolson steps.
    These scheme are second order accurate for the implicit terms, but potentially
    higher order accurate for the explicit terms. This seems to be a favorable
    tradeoff when the explicit terms dominate, e.g., for modeling turbulent
    fluids.
    Per Canuto: "[these methods] have been widely used for the time-discretization
    in applications of spectral methods."
    Args:
      alphas: alpha coefficients.
      betas: beta coefficients.
      gammas: gamma coefficients.
      equation: equation to solve.
      time_step: time step.
    Returns:
      Function that performs a time step.
    Reference:
      Canuto, C., Yousuff Hussaini, M., Quarteroni, A. & Zang, T. A.
      Spectral Methods: Evolution to Complex Geometries and Applications to
      Fluid Dynamics. (Springer Berlin Heidelberg, 2007).
      https://doi.org/10.1007/978-3-540-30728-0 (Appendix D.3)
    """
    # pylint: disable=invalid-name,non-ascii-name
    α = alphas
    β = betas
    γ = gammas
    dt = time_step
    F = tree_math.unwrap(equation.explicit_terms)
    G = tree_math.unwrap(equation.implicit_terms)
    G_inv = tree_math.unwrap(equation.implicit_solve, vector_argnums=0)

    if len(alphas) - 1 != len(betas) != len(gammas):
        raise ValueError("number of RK coefficients does not match")

    @tree_math.wrap
    def step_fn(u):
        h = 0
        for k in range(len(β)):
            h = F(u) + β[k] * h
            µ = 0.5 * dt * (α[k + 1] - α[k])
            u = G_inv(u + γ[k] * dt * h + µ * G(u), µ)
        return u

    return step_fn


def low_storage_runge_kutta_crank_nicolson_cde(
    alphas: Sequence[float],
    betas: Sequence[float],
    gammas: Sequence[float],
    equation: ImplicitExplicitCDE,
    time_step: float,
) -> ControlledTimeStepFn:
    """Time stepping via "low-storage" Runge-Kutta and Crank-Nicolson steps.
    These scheme are second order accurate for the implicit terms, but potentially
    higher order accurate for the explicit terms. This seems to be a favorable
    tradeoff when the explicit terms dominate, e.g., for modeling turbulent
    fluids.
    Per Canuto: "[these methods] have been widely used for the time-discretization
    in applications of spectral methods."
    Args:
      alphas: alpha coefficients.
      betas: beta coefficients.
      gammas: gamma coefficients.
      equation: equation to solve.
      time_step: time step.
    Returns:
      Function that performs a time step.
    Reference:
      Canuto, C., Yousuff Hussaini, M., Quarteroni, A. & Zang, T. A.
      Spectral Methods: Evolution to Complex Geometries and Applications to
      Fluid Dynamics. (Springer Berlin Heidelberg, 2007).
      https://doi.org/10.1007/978-3-540-30728-0 (Appendix D.3)
    """
    # pylint: disable=invalid-name,non-ascii-name
    α = alphas
    β = betas
    γ = gammas
    dt = time_step
    F = tree_math.unwrap(equation.explicit_terms)
    G = tree_math.unwrap(equation.implicit_terms)
    G_inv = tree_math.unwrap(equation.implicit_solve, vector_argnums=0)

    if len(alphas) - 1 != len(betas) != len(gammas):
        raise ValueError("number of RK coefficients does not match")

    @tree_math.wrap
    def step_fn(u, x):
        h = 0
        for k in range(len(β)):
            h = F(u, x) + β[k] * h
            µ = 0.5 * dt * (α[k + 1] - α[k])
            u = G_inv(u + γ[k] * dt * h + µ * G(u), µ)
        return u, u

    return step_fn


def crank_nicolson_rk3(
    equation: ImplicitExplicitODE,
    time_step: float,
) -> TimeStepFn:
    """Time stepping via Crank-Nicolson and RK3 ("Williamson")."""
    return low_storage_runge_kutta_crank_nicolson(
        alphas=[0, 1 / 3, 3 / 4, 1],
        betas=[0, -5 / 9, -153 / 128],
        gammas=[1 / 3, 15 / 16, 8 / 15],
        equation=equation,
        time_step=time_step,
    )


def crank_nicolson_rk3_cde(
    equation: ImplicitExplicitCDE,
    time_step: float,
) -> ControlledTimeStepFn:
    """Time stepping via Crank-Nicolson and RK3 ("Williamson")."""
    return low_storage_runge_kutta_crank_nicolson_cde(
        alphas=[0, 1 / 3, 3 / 4, 1],
        betas=[0, -5 / 9, -153 / 128],
        gammas=[1 / 3, 15 / 16, 8 / 15],
        equation=equation,
        time_step=time_step,
    )


def crank_nicolson_rk4(
    equation: ImplicitExplicitODE,
    time_step: float,
) -> TimeStepFn:
    """Time stepping via Crank-Nicolson and RK4 ("Carpenter-Kennedy")."""
    # pylint: disable=line-too-long
    return low_storage_runge_kutta_crank_nicolson(
        alphas=[
            0,
            0.1496590219993,
            0.3704009573644,
            0.6222557631345,
            0.9582821306748,
            1,
        ],
        betas=[0, -0.4178904745, -1.192151694643, -1.697784692471, -1.514183444257],
        gammas=[
            0.1496590219993,
            0.3792103129999,
            0.8229550293869,
            0.6994504559488,
            0.1530572479681,
        ],
        equation=equation,
        time_step=time_step,
    )


def crank_nicolson_rk4_cde(
    equation: ImplicitExplicitCDE,
    time_step: float,
) -> ControlledTimeStepFn:
    """Time stepping via Crank-Nicolson and RK4 ("Carpenter-Kennedy")."""
    # pylint: disable=line-too-long
    return low_storage_runge_kutta_crank_nicolson_cde(
        alphas=[
            0,
            0.1496590219993,
            0.3704009573644,
            0.6222557631345,
            0.9582821306748,
            1,
        ],
        betas=[0, -0.4178904745, -1.192151694643, -1.697784692471, -1.514183444257],
        gammas=[
            0.1496590219993,
            0.3792103129999,
            0.8229550293869,
            0.6994504559488,
            0.1530572479681,
        ],
        equation=equation,
        time_step=time_step,
    )


@dataclasses.dataclass
class ImExButcherTableau:
    """Butcher Tableau for implicit-explicit Runge-Kutta methods."""

    a_ex: Sequence[Sequence[float]]
    a_im: Sequence[Sequence[float]]
    b_ex: Sequence[float]
    b_im: Sequence[float]

    def __post_init__(self):
        if (
            len(
                {len(self.a_ex) + 1, len(self.a_im) + 1, len(self.b_ex), len(self.b_im)}
            )
            > 1
        ):
            raise ValueError("inconsistent Butcher tableau")


def imex_runge_kutta(
    tableau: ImExButcherTableau,
    equation: ImplicitExplicitODE,
    time_step: float,
) -> TimeStepFn:
    """Time stepping with Implicit-Explicit Runge-Kutta."""
    # pylint: disable=invalid-name
    dt = time_step
    F = tree_math.unwrap(equation.explicit_terms)
    G = tree_math.unwrap(equation.implicit_terms)
    G_inv = tree_math.unwrap(equation.implicit_solve, vector_argnums=0)

    a_ex = tableau.a_ex
    a_im = tableau.a_im
    b_ex = tableau.b_ex
    b_im = tableau.b_im

    num_steps = len(b_ex)

    @tree_math.wrap
    def step_fn(y0):
        f = [None] * num_steps
        g = [None] * num_steps

        f[0] = F(y0)
        g[0] = G(y0)

        for i in range(1, num_steps):
            ex_terms = dt * sum(
                a_ex[i - 1][j] * f[j] for j in range(i) if a_ex[i - 1][j]
            )
            im_terms = dt * sum(
                a_im[i - 1][j] * g[j] for j in range(i) if a_im[i - 1][j]
            )
            Y_star = y0 + ex_terms + im_terms
            Y = G_inv(Y_star, dt * a_im[i - 1][i])
            if any(a_ex[j][i] for j in range(i, num_steps - 1)) or b_ex[i]:
                f[i] = F(Y)
            if any(a_im[j][i] for j in range(i, num_steps - 1)) or b_im[i]:
                g[i] = G(Y)

        ex_terms = dt * sum(b_ex[j] * f[j] for j in range(num_steps) if b_ex[j])
        im_terms = dt * sum(b_im[j] * g[j] for j in range(num_steps) if b_im[j])
        y_next = y0 + ex_terms + im_terms

        return y_next

    return step_fn


def imex_runge_kutta_cde(
    tableau: ImExButcherTableau,
    equation: ImplicitExplicitCDE,
    time_step: float,
) -> ControlledTimeStepFn:
    """Time stepping with Implicit-Explicit Runge-Kutta."""
    # pylint: disable=invalid-name
    dt = time_step
    F = tree_math.unwrap(equation.explicit_terms)
    G = tree_math.unwrap(equation.implicit_terms)
    G_inv = tree_math.unwrap(equation.implicit_solve, vector_argnums=0)

    a_ex = tableau.a_ex
    a_im = tableau.a_im
    b_ex = tableau.b_ex
    b_im = tableau.b_im

    num_steps = len(b_ex)

    @tree_math.wrap
    def step_fn(y0, x0):
        f = [None] * num_steps
        g = [None] * num_steps

        f[0] = F(y0, x0)
        g[0] = G(y0)

        for i in range(1, num_steps):
            ex_terms = dt * sum(
                a_ex[i - 1][j] * f[j] for j in range(i) if a_ex[i - 1][j]
            )
            im_terms = dt * sum(
                a_im[i - 1][j] * g[j] for j in range(i) if a_im[i - 1][j]
            )
            Y_star = y0 + ex_terms + im_terms
            Y = G_inv(Y_star, dt * a_im[i - 1][i])
            if any(a_ex[j][i] for j in range(i, num_steps - 1)) or b_ex[i]:
                f[i] = F(Y, x0)
            if any(a_im[j][i] for j in range(i, num_steps - 1)) or b_im[i]:
                g[i] = G(Y)

        ex_terms = dt * sum(b_ex[j] * f[j] for j in range(num_steps) if b_ex[j])
        im_terms = dt * sum(b_im[j] * g[j] for j in range(num_steps) if b_im[j])
        y_next = y0 + ex_terms + im_terms

        return y_next, y_next

    return step_fn


def imex_rk_sil3(
    equation: ImplicitExplicitODE,
    time_step: float,
) -> TimeStepFn:
    """Time stepping with the SIL3 implicit-explicit RK scheme.
    This method is second-order accurate for the implicit terms and third-order
    accurate for the explicit terms.
    Args:
      equation: equation to solve.
      time_step: time step.
    Returns:
      Function that performs a time step.
    Reference:
      Whitaker, J. S. & Kar, S. K. Implicit-Explicit Runge-Kutta Methods for
      Fast-Slow Wave Problems. Monthly Weather Review vol. 141 3426-3434 (2013)
      http://dx.doi.org/10.1175/mwr-d-13-00132.1
    """
    return imex_runge_kutta(
        tableau=ImExButcherTableau(
            a_ex=[[1 / 3], [1 / 6, 1 / 2], [1 / 2, -1 / 2, 1]],
            a_im=[[1 / 6, 1 / 6], [1 / 3, 0, 1 / 3], [3 / 8, 0, 3 / 8, 1 / 4]],
            b_ex=[1 / 2, -1 / 2, 1, 0],
            b_im=[3 / 8, 0, 3 / 8, 1 / 4],
        ),
        equation=equation,
        time_step=time_step,
    )


def imex_rk_sil3_cde(
    equation: ImplicitExplicitCDE,
    time_step: float,
) -> ControlledTimeStepFn:
    """Time stepping with the SIL3 implicit-explicit RK scheme.
    This method is second-order accurate for the implicit terms and third-order
    accurate for the explicit terms.
    Args:
      equation: equation to solve.
      time_step: time step.
    Returns:
      Function that performs a time step.
    Reference:
      Whitaker, J. S. & Kar, S. K. Implicit-Explicit Runge-Kutta Methods for
      Fast-Slow Wave Problems. Monthly Weather Review vol. 141 3426-3434 (2013)
      http://dx.doi.org/10.1175/mwr-d-13-00132.1
    """
    return imex_runge_kutta_cde(
        tableau=ImExButcherTableau(
            a_ex=[[1 / 3], [1 / 6, 1 / 2], [1 / 2, -1 / 2, 1]],
            a_im=[[1 / 6, 1 / 6], [1 / 3, 0, 1 / 3], [3 / 8, 0, 3 / 8, 1 / 4]],
            b_ex=[1 / 2, -1 / 2, 1, 0],
            b_im=[3 / 8, 0, 3 / 8, 1 / 4],
        ),
        equation=equation,
        time_step=time_step,
    )



# Copyright (c) 2022 Heidelberg University. All rights reserved.
#
# Released under Apache 2.0 license as described in the file LICENSE.
# Authors: Christian Pehle

import arbor
import jax.numpy as jnp
import numpy as onp
from absl.testing import absltest

import njx.morph.generate_morphologies as gm
from njx.base.tree_solver import hines_solver, tree_matmul, tree_to_matrix


def test_tree_to_matrix():
    N = 3
    d = 2 * jnp.ones(N)
    u = jnp.ones(N - 1)
    p = jnp.arange(-1, N, 1)

    expected = [[2, 1, 0], [1, 2, 1], [0, 1, 2]]
    actual = tree_to_matrix(d, u, p)
    onp.testing.assert_allclose(expected, actual, rtol=1e-4)


def test_tree_to_matrix_2():
    morph = gm.y_geometry()
    tm = gm.compute_tree_matrix(morph)

    expected = [[-2, 1, 0, 0], [1, -3, 1, 1], [0, 1, -2, 0], [0, 1, 0, -2]]

    actual = tree_to_matrix(tm.d, tm.u, tm.p)
    onp.testing.assert_allclose(expected, actual)


def test_hines_solver():
    N = 10
    d = 2 * onp.random.randn(N)
    u = onp.random.randn(N - 1)
    b = onp.random.randn(N)
    p = onp.arange(-1, N, 1)
    a = onp.diag(d, 0) + onp.diag(u, 1) + onp.diag(u, -1)

    x = hines_solver(jnp.array(d), jnp.array(u), jnp.array(p), jnp.array(b))
    x_ = jnp.linalg.solve(a, b)

    onp.testing.assert_allclose(x, x_, rtol=1e-4)


def test_hines_solver_2():
    morph = gm.y_geometry()
    tm = gm.compute_tree_matrix(morph)
    b = onp.random.randn(4)
    a = tree_to_matrix(tm.d + 0.01, tm.u, tm.p)

    x = hines_solver(tm.d + 0.01, tm.u, tm.p, jnp.array(b))
    x_ = jnp.linalg.solve(a, jnp.array(b))

    # TODO: This is a rather liberal error tolerance...
    onp.testing.assert_allclose(x, x_, rtol=1e-4)


def test_hines_solver_3():
    morph = gm.y_geometry()
    tm = gm.compute_tree_matrix(morph)
    b = onp.random.randn(4)

    # TODO: why is the 0.01 needed!?
    a = tree_to_matrix(tm.d + 0.01, tm.u, tm.p)
    x = hines_solver(tm.d + 0.01, tm.u, tm.p, jnp.array(b))
    x_ = jnp.linalg.solve(a, jnp.array(b))

    # TODO: This is a rather liberal error tolerance...
    onp.testing.assert_allclose(x, x_, rtol=1e-4)


def test_hines_solver_4():
    morph = gm.branched_geometry()
    tm = gm.compute_tree_matrix(morph)
    b = onp.random.randn(tm.d.shape[0])

    # TODO: why is the 0.01 needed!?
    a = tree_to_matrix(tm.d + 0.01, tm.u, tm.p)
    x = hines_solver(tm.d + 0.01, tm.u, tm.p, jnp.array(b))
    x_ = jnp.linalg.solve(a, jnp.array(b))

    # TODO: This is a rather liberal error tolerance...
    onp.testing.assert_allclose(x, x_, rtol=1e-2)


def test_hines_solver_5():
    loaded_morphology = gm.swc_geometry(
        "data/morphologies/allen/Cux2-CreERT2_Ai14-211772.05.02.01_674408996_m.swc"
    )
    morph = loaded_morphology.morphology
    tm = gm.compute_tree_matrix(morph, policy=arbor.cv_policy_fixed_per_branch(1))
    b = onp.random.randn(tm.d.shape[0])

    # TODO: why is the 0.01 needed!?
    a = tree_to_matrix(tm.d + 0.01, tm.u, tm.p)
    x = hines_solver(tm.d + 0.01, tm.u, tm.p, jnp.array(b))
    x_ = jnp.linalg.solve(a, jnp.array(b))

    # TODO: This is a rather liberal error tolerance...
    onp.testing.assert_allclose(x, x_, rtol=1e-1)


def test_hines_solver_6():
    loaded_morphology = gm.swc_geometry(
        "data/morphologies/allen/Cux2-CreERT2_Ai14-211772.05.02.01_674408996_m.swc"
    )
    morph = loaded_morphology.morphology
    tm = gm.compute_tree_matrix(morph, policy=arbor.cv_policy_fixed_per_branch(3))
    b = onp.random.randn(tm.d.shape[0])

    # TODO: why is the 0.01 needed!?
    a = tree_to_matrix(tm.d + 0.01, tm.u, tm.p)
    x = hines_solver(tm.d + 0.01, tm.u, tm.p, jnp.array(b))
    x_ = jnp.linalg.solve(a, jnp.array(b))

    # TODO: This is a rather liberal error tolerance...
    onp.testing.assert_allclose(x, x_, rtol=1e-2)


def test_tree_matmul():
    N = 3
    d = jnp.array(2 * onp.ones(N))
    u = jnp.array(onp.ones(N - 1))
    p = jnp.array(onp.arange(-1, N, 1))

    mat = tree_to_matrix(d, u, p)
    a = jnp.array(onp.random.randn(3))
    expected = jnp.dot(mat, a)
    actual = tree_matmul(d, u, p, a)

    onp.testing.assert_allclose(expected, actual)


def test_tree_matmul_2():
    morph = gm.y_geometry()
    tm = gm.compute_tree_matrix(morph)
    a = onp.random.randn(4)
    mat = tree_to_matrix(tm.d, tm.u, tm.p)

    expected = jnp.dot(mat, a)
    actual = tree_matmul(tm.d, tm.u, tm.p, a)

    onp.testing.assert_allclose(expected, actual, rtol=0.1)


if __name__ == "__main__":
    absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from typing import Callable, Sequence, TypeVar
import jax
import tree_math


PyTreeState = TypeVar("PyTreeState")
TimeStepFn = Callable[[PyTreeState], PyTreeState]


class ExplicitConstrainedODE:
    """

    The equation is given by:
      ∂u/∂t = explicit_terms(u)
      0 = constraint(u)
    """

    def __init__(self, explicit_terms, projection):
        self.explicit_terms = explicit_terms
        self.projection = projection

    def explicit_terms(self, state):
        """Explicitly evaluate the ODE."""
        raise NotImplementedError

    def projection(self, state):
        """Enforce the constraint."""
        raise NotImplementedError


class ExplicitConstrainedCDE:
    """

    The equation is given by:
      ∂u/∂t = explicit_terms(u, x)
      0 = constraint(u)
    """

    def __init__(self, explicit_terms, projection, output):
        self.explicit_terms = explicit_terms
        self.projection = projection
        self.output = output

    def explicit_terms(self, state, input):
        """Explicitly evaluate the ODE."""
        raise NotImplementedError

    def projection(self, state, input):
        """Enforce the constraint."""
        raise NotImplementedError

    def output(self, state):
        """"""
        raise NotImplementedError


@dataclasses.dataclass
class ButcherTableau:
    a: Sequence[Sequence[float]]
    b: Sequence[float]
    # TODO(shoyer): add c, when we support time-dependent equations.

    def __post_init__(self):
        if len(self.a) + 1 != len(self.b):
            raise ValueError("inconsistent Butcher tableau")


def constrained_rk(
    tableau: ButcherTableau,
    equation: ExplicitConstrainedODE,
    time_step: float,
) -> TimeStepFn:
    """Create a forward Runge-Kutta time-stepper for incompressible Navier-Stokes.
    This function implements the reference method (equations 16-21), rather than
    the fast projection method, from:
      "Fast-Projection Methods for the Incompressible Navier–Stokes Equations"
      Fluids 2020, 5, 222; doi:10.3390/fluids5040222
    Args:
      tableau: Butcher tableau.
      equation: equation to use.
      time_step: overall time-step size.
    Returns:
      Function that advances one time-step forward.
    """
    # pylint: disable=invalid-name
    dt = time_step
    F = tree_math.unwrap(equation.explicit_terms)
    P = tree_math.unwrap(equation.projection)

    a = tableau.a
    b = tableau.b
    num_steps = len(b)

    @tree_math.wrap
    def step_fn(u0):
        u = [None] * num_steps
        k = [None] * num_steps

        u[0] = u0
        k[0] = F(u0)

        for i in range(1, num_steps):
            u_star = u0 + dt * sum(a[i - 1][j] * k[j] for j in range(i) if a[i - 1][j])
            u[i] = P(u_star)
            k[i] = F(u[i])

        u_star = u0 + dt * sum(b[j] * k[j] for j in range(num_steps) if b[j])
        u_final = P(u_star)

        return u_final

    return step_fn


def constrained_cde_rk(
    tableau: ButcherTableau,
    equation: ExplicitConstrainedCDE,
    time_step: float,
) -> TimeStepFn:
    """Create a forward Runge-Kutta time-stepper for incompressible Navier-Stokes.
    This function implements the reference method (equations 16-21), rather than
    the fast projection method, from:
      "Fast-Projection Methods for the Incompressible Navier–Stokes Equations"
      Fluids 2020, 5, 222; doi:10.3390/fluids5040222
    Args:
      tableau: Butcher tableau.
      equation: equation to use.
      time_step: overall time-step size.
    Returns:
      Function that advances one time-step forward.
    """
    # pylint: disable=invalid-name
    dt = time_step
    F = tree_math.unwrap(equation.explicit_terms)
    P = tree_math.unwrap(equation.projection)
    O = tree_math.unwrap(equation.output)

    a = tableau.a
    b = tableau.b
    num_steps = len(b)

    @tree_math.wrap
    def step_fn(u0, x0):
        u = [None] * num_steps
        k = [None] * num_steps

        u[0] = u0
        k[0] = F(u0, x0)

        for i in range(1, num_steps):
            u_star = u0 + dt * sum(a[i - 1][j] * k[j] for j in range(i) if a[i - 1][j])
            u[i] = P(u_star, x0)
            k[i] = F(u[i], x0)

        u_star = u0 + dt * sum(b[j] * k[j] for j in range(num_steps) if b[j])

        out = O(u_star, x0)
        u_final = P(u_star, x0)

        return u_final, out

    return step_fn


def forward_euler(
    equation: ExplicitConstrainedODE,
    time_step: float,
) -> TimeStepFn:
    return jax.named_call(
        constrained_rk(ButcherTableau(a=[], b=[1]), equation, time_step),
        name="forward_euler",
    )


def forward_euler_cde(
    equation: ExplicitConstrainedCDE,
    time_step: float,
) -> TimeStepFn:
    return jax.named_call(
        constrained_cde_rk(ButcherTableau(a=[], b=[1]), equation, time_step),
        name="forward_euler",
    )


def midpoint_rk2(
    equation: ExplicitConstrainedODE,
    time_step: float,
) -> TimeStepFn:
    return jax.named_call(
        constrained_rk(
            ButcherTableau(a=[[1 / 2]], b=[0, 1]),
            equation=equation,
            time_step=time_step,
        ),
        name="midpoint_rk2",
    )


def midpoint_rk2_cde(
    equation: ExplicitConstrainedCDE,
    time_step: float,
) -> TimeStepFn:
    return jax.named_call(
        constrained_cde_rk(
            ButcherTableau(a=[[1 / 2]], b=[0, 1]),
            equation=equation,
            time_step=time_step,
        ),
        name="midpoint_rk2",
    )


def heun_rk2(
    equation: ExplicitConstrainedODE,
    time_step: float,
) -> TimeStepFn:
    return jax.named_call(
        constrained_rk(
            ButcherTableau(a=[[1]], b=[1 / 2, 1 / 2]),
            equation=equation,
            time_step=time_step,
        ),
        name="heun_rk2",
    )


def heun_rk2_cde(
    equation: ExplicitConstrainedCDE,
    time_step: float,
) -> TimeStepFn:
    return jax.named_call(
        constrained_cde_rk(
            ButcherTableau(a=[[1]], b=[1 / 2, 1 / 2]),
            equation=equation,
            time_step=time_step,
        ),
        name="heun_rk2",
    )


def classic_rk4(
    equation: ExplicitConstrainedODE,
    time_step: float,
) -> TimeStepFn:
    return jax.named_call(
        constrained_rk(
            ButcherTableau(
                a=[[1 / 2], [0, 1 / 2], [0, 0, 1]], b=[1 / 6, 1 / 3, 1 / 3, 1 / 6]
            ),
            equation=equation,
            time_step=time_step,
        ),
        name="classic_rk4",
    )


def classic_rk4_cde(
    equation: ExplicitConstrainedCDE,
    time_step: float,
) -> TimeStepFn:
    return jax.named_call(
        constrained_cde_rk(
            ButcherTableau(
                a=[[1 / 2], [0, 1 / 2], [0, 0, 1]], b=[1 / 6, 1 / 3, 1 / 3, 1 / 6]
            ),
            equation=equation,
            time_step=time_step,
        ),
        name="classic_rk4",
    )



# Copyright (c) 2022 Heidelberg University. All rights reserved.
#
# Released under Apache 2.0 license as described in the file LICENSE.
# Authors: Christian Pehle

import math

def channel_time_constant(alpha, beta, v_shift=0.0, q=1.0):
    def tau(v):
        v = v + v_shift
        return 1 / q * 1 / (alpha(v) + beta(v))

    return tau


def channel_equilibrium(alpha, beta, v_shift=0.0):
    def x_inf(v):
        v = v + v_shift
        return alpha(v) / (alpha(v) + beta(v))

    return x_inf


def q_conversion(target_deg_c, original_deg_c, q10):
    return math.exp(q10, (target_deg_c - original_deg_c) / 10)


def channel_dynamics(alpha, beta):
    def dynamics(x, voltage):
        return alpha(voltage) * (1 - x) - beta(voltage) * x
    return dynamics


# Copyright (c) 2022 Heidelberg University. All rights reserved.
#
# Released under Apache 2.0 license as described in the file LICENSE.
# Authors: Christian Pehle

# This is a port of the "Traub.mod" found in https://github.com/ModelDBRepository/254217/_mod/Traub.mod by Albert Gidon & Leora Menhaim (2004).
#
# Some original comments are reproduced below with the prefix ":".
# Please refer to the original file for additional attributions.
#
# References:
#
# Gidon A, Zolnik TA, Fidzinski P, Bolduan F, Papoutsi A, Poirazi P, Holtkamp M, Vida I, Larkum ME (2020).
#   Dendritic action potentials and computation in human layer 2/3 cortical neurons
#   Science 367:83-87
#
# Traub, R. D., Wong, R. K., Miles, R., and Michelson, H. (1991).
# 	A model of a CA3 hippocampal pyramidal neuron incorporating
# 	voltage-clamp data on intrinsic conductances.
# 	J Neurophysiol 66, 635-650.
#
# Kang, S., Kitano, K., and Fukai, T. (2004).
# 	Self-organized two-state membrane potential
# 	transitions in a network of realistically modeled
# 	cortical neurons. Neural Netw 17, 307-312.

import dataclasses

import jax.numpy as jnp
import tree_math
from njx.base.types import ArrayLike

from njx.channels.common import (
    channel_equilibrium,
    channel_time_constant,
    q_conversion,
)


@dataclasses.dataclass
@tree_math.struct
class TraubParameters:
    g_Na: ArrayLike = 0.03  # (S/cm2)	:Traub et. al. 1991
    g_K: ArrayLike = 0.015  # (S/cm2) :Traub et. al. 1991
    g_L: ArrayLike = 0.00014  # (S/cm2) :Siu Kang - by email.
    e_L: ArrayLike = -62.0  # (mV)    :Siu Kang - by email.
    e_K: ArrayLike = -80  # (mV)    :Siu Kang - by email.
    e_Na: ArrayLike = 90  # (mV)    :Leora


# e_L = -74.0
# e_K = -80
# e_Na = 90

# magic constants
_v_shift = 49.2  # shift to apply to all curves
_q_10 = 3
_target_temperature = 35
_original_temperature = 32

# temperature conversion
_q = q_conversion(_target_temperature, _original_temperature, _q_10)


@dataclasses.dataclass
@tree_math.struct
class TraubState:
    m: ArrayLike
    h: ArrayLike
    n: ArrayLike


def alpha_h(v: ArrayLike):
    return 0.128 * jnp.exp((17 - v) / 18)


def beta_h(v: ArrayLike):
    return 4 / (1 + jnp.exp((40 - v) / 5))


tau_h = channel_time_constant(alpha=alpha_h, beta=beta_h, v_shift=_v_shift, q=_q)
h_inf = channel_equilibrium(alpha=alpha_h, beta=beta_h)


def alpha_m(v):
    return jnp.where(
        v == 13.1, 0.32 * 4, 0.32 * (13.1 - v) / (jnp.exp((13.1 - v) / 4) - 1)
    )


def beta_m(v):
    return jnp.where(
        v == 40.1, 0.28 * 5, 0.28 * (v - 40.1) / (jnp.exp((v - 40.1) / 5) - 1)
    )


tau_m = channel_time_constant(alpha=alpha_m, beta=beta_m, v_shift=_v_shift, q=_q)
m_inf = channel_equilibrium(alpha=alpha_m, beta=beta_m, v_shift=_v_shift)


def alpha_n(v):
    return jnp.where(
        v == 35.1, 0.016 * 5, 0.016 * (35.1 - v) / (jnp.exp((35.1 - v) / 5) - 1)
    )


def beta_n(v):
    return 0.25 * jnp.exp((20 - v) / 40)


tau_n = channel_time_constant(alpha=alpha_n, beta=beta_n, v_shift=_v_shift, q=_q)
n_inf = channel_equilibrium(alpha=alpha_n, beta=beta_n, v_shift=_v_shift)


def g_Na(s: TraubState, p: TraubParameters):
    return p.g_Na * s.h * s.m * s.m


def I_Na(v: ArrayLike, s: TraubState, p: TraubParameters):
    return g_Na(s, p) * (v - p.e_Na)


def g_K(s: TraubState, p: TraubParameters):
    return p.g_K * s.n


def I_K(v: ArrayLike, s: TraubState, p: TraubParameters):
    return g_K(s, p) * (v - p.e_K)


def I_L(v: ArrayLike, _: TraubState, p: TraubParameters):
    return p.g_L * (v - p.e_L)


def dynamics(v: ArrayLike, s: TraubState, p: TraubParameters):
    return TraubState(
        m=(m_inf(v) - s.m) / tau_m(v),
        h=(h_inf(v) - s.h) / tau_h(v),
        n=(n_inf(v) - s.n) / tau_n(v),  # :phi=12 from Kang et. al. 2004
    )



# Copyright (c) 2022 Heidelberg University. All rights reserved.
#
# Released under Apache 2.0 license as described in the file LICENSE.
# Authors: Christian Pehle

# This is a port of the the LVA ca channel found in https://github.com/ModelDBRepository/254217/ by Albert Gidon & Leora Menhaim (2004).
# Some of the original comments are retained and prefixed with ":"
#
# : LVA ca channel. Note: mtau is an approximation from the plots
# : Reference: Avery and Johnston 1996, tau from Randall 1997
# : shifted by -10 mv to correct for junction potential
# : corrected rates using q10 = 2.3, target temperature 34, orginal 21

import dataclasses

import jax.numpy as jnp
import tree_math

from njx.base.types import ArrayLike


@dataclasses.dataclass
@tree_math.struct
class CaLVAState:
    m: ArrayLike
    h: ArrayLike


@tree_math.struct
class CaLVAParameters:
    g_bar = 0.00001  # (S/cm2)


def g(s: CaLVAState, p: CaLVAParameters):
    return p.g_bar * s.m * s.m * s.h


def I_Ca(v: ArrayLike, e_ca: ArrayLike, s: CaLVAState, p: CaLVAParameters):
    return g(s, p) * (v - e_ca)


def m_inf(v):
    v = v + 10
    return 1.0000 / (1 + jnp.exp((v - -30.000) / -6))


def m_tau(v):
    v = v + 10
    celsius = 34
    qt = 2.3 ** ((celsius - 21) / 10)  # TODO: Find better way to adjust for temperature
    return (5.0000 + 20.0000 / (1 + jnp.exp((v - -25.000) / 5))) / qt


def h_inf(v):
    v = v + 10
    return 1.0000 / (1 + jnp.exp((v - -80.000) / 6.4))


def h_tau(v):
    v = v + 10
    celsius = 34
    qt = 2.3 ** ((celsius - 21) / 10)  # TODO: Find better way to adjust for temperature
    return (20.0000 + 50.0000 / (1 + jnp.exp((v - -40.000) / 7))) / qt


def dynamics(s: CaLVAState, v: ArrayLike):
    return CaLVAState(m=(m_inf(v) - s.m) / m_tau(v), h=(h_inf(v) - s.h) / h_tau(v))



# Copyright (c) 2022 Heidelberg University. All rights reserved.
#
# Released under Apache 2.0 license as described in the file LICENSE.
# Authors: Christian Pehle


# This is in part a port of "ISyn.inc" found in https://github.com/ModelDBRepository/254217/_mod/ISyn.inc
# and also incorporates the parameters found in https://github.com/ModelDBRepository/254217/Fig3andS9/syns.hoc
# Some of the original comments are preserved with prefix ":"

import dataclasses

import jax.numpy as jnp
import tree_math
from njx.base.types import ArrayLike


@dataclasses.dataclass
@tree_math.struct
class SynapseState:
    a: ArrayLike
    b: ArrayLike


@dataclasses.dataclass
@tree_math.struct
class SynapseParameters:
    tau_r: ArrayLike
    tau_d: ArrayLike
    g_max: ArrayLike
    gamma: ArrayLike = 0.062
    e: ArrayLike = 0.0
    mg: ArrayLike = 0.0
    mg_dep: ArrayLike = 3.57 #  Jahr & Stevens 1990

# mggate kinetics - Jahr & Stevens 1990

# mggate = 1 / (1 + exp(0.08  (/mV) * -(v)) * (mg / 3.57 (mM)))


def weight_factor(tau_fast, tau_slow):
    t = (tau_fast * tau_slow) / (tau_slow - tau_fast) * jnp.log(tau_slow / tau_fast)
    e = -jnp.exp(-t / tau_fast) + jnp.exp(-t / tau_slow)
    return 1 / e


def dynamics(s: SynapseState, p: SynapseParameters):
    return SynapseState(a=-1 / p.tau_r * s.a, b=-1 / p.tau_d * s.b)


def voltage_gated_dynamics(gating_function):
    def dyn(s: SynapseState, v: ArrayLike, v_dot: ArrayLike, p: SynapseParameters):
        return SynapseState(
            a=-1 / p.tau_r * s.a
            + gating_function(v) * weight_factor(p.tau_r, p.tau_d) * v_dot,
            b=-1 / p.tau_d * s.b
            + gating_function(v) * weight_factor(p.tau_r, p.tau_d) * v_dot,
        )

    return dyn


def transition(w: ArrayLike, s: SynapseState, p: SynapseParameters):
    return SynapseState(
        a=s.a + w * weight_factor(p.tau_r, p.tau_d),
        b=s.b + w * weight_factor(p.tau_r, p.tau_d),
    )


def g_syn(s: SynapseState, p: SynapseParameters):
    return p.g_max * (s.b - s.a)


def mg_dependence(v: ArrayLike, p: SynapseParameters):
    return 1 / (1 + jnp.exp(-p.gamma * v) * (p.mg / p.mg_dep))


def I_syn(v: ArrayLike, s: SynapseState, p: SynapseParameters):
    return g_syn(s, p) * mg_dependence(v, p) * (v - p.e)


# :  Synaptic parameters as in Eyal et al., (2018)
# : AMPA
ampa_parameters = SynapseParameters(g_max=0.7e-3, tau_r=0.3, tau_d=1.8, mg=0, e=0)  # uS

# : NMDA
nmda_parameters = SynapseParameters(
    g_max=1.3e-3,  # uS
    tau_r=8,
    tau_d=35,
    mg=1,
    e=0,
    gamma=0.077 / 0.082,  # : rhodes et al., 2006
)

# : GABA
# : mid value between fast GABAa with 5 ms decay as in Salin and Prince, 1996 for
# : fast basket cells inhibition and slow martinotti GABAa with 20 ms decay (Gidon and Segev 2012)
basket_gaba_parameters = SynapseParameters(
    g_max=0.5e-3, tau_r=0.5, tau_d=5.0, mg=0, e=-75  # uS
)

martinotti_gaba_parameters = SynapseParameters(
    g_max=0.5e-3, tau_r=2.0, tau_d=23.0, mg=0.0, e=-75  # uS
)



# Copyright (c) 2022 Heidelberg University. All rights reserved.
#
# Released under Apache 2.0 license as described in the file LICENSE.
# Authors: Christian Pehle

from absl.testing import absltest
from absl.testing import parameterized

import numpy as onp
import jax.numpy as np
import jax
import math

import njx.base.root_solving as root_solving


ALL_TEST_PROBLEMS = [
    dict(
        testcase_name="root of unity",
        f=lambda x: x**20 - 1,
        df=lambda x: 20 * x**19,
        bounds=[(0.0, 4.0)],
        roots=[-1.0, 1.0],
    ),
    dict(
        testcase_name="sign",
        f=lambda x: np.sqrt(np.abs(x)) * np.sign(x),
        df=lambda x: 1 / np.sqrt(np.abs(x)),
        bounds=[(-1.0, 1.0)],
        roots=[0.0],
    ),
    dict(
        testcase_name="squareroot",
        f=lambda x: x**2 - 2,
        df=lambda x: 2 * x,
        bounds=[(0.0, 2.0)],
        roots=[-math.sqrt(2), math.sqrt(2)],
    ),
    dict(
        testcase_name="closeroot",
        f=lambda x: x**2 - 1e-8,
        df=lambda x: 2 * x,
        bounds=[(0.0, 2.0)],
        roots=[-math.sqrt(1e-8), math.sqrt(1e-8)],
    ),
    dict(
        testcase_name="xexp",
        f=lambda x: x * np.exp(-x),
        df=lambda x: np.exp(-x) - x * np.exp(-x),
        bounds=[(-1.0, 1.0)],
        roots=[0.0],
    ),
    # TODO: Doesn't work!
    # dict(
    #     testcase_name="xpow7",
    #     f=lambda x: (x - 1)**7,
    #     df=lambda x: 7*(x-1)**6,
    #     bounds = [(0.0,2.0)],
    #     roots=[1.0]
    # )
]

ALL_BOUNDED_ROOT_METHODS = [root_solving.bisection, root_solving.illinois_method]

ALL_GRADIENT_BASED_METHODS = [root_solving.newton_1d]


class RootSolvingTest(parameterized.TestCase):
    @parameterized.named_parameters(ALL_TEST_PROBLEMS)
    def test_root_solving(self, f, df, bounds, roots):
        bound = bounds[0]
        expected = roots[-1]
        tol = 0.01
        for solver in ALL_BOUNDED_ROOT_METHODS:
            actual = solver(f, bound[0], bound[1], tol)
            onp.testing.assert_allclose(expected, actual, atol=tol)

    @parameterized.named_parameters(ALL_TEST_PROBLEMS)
    def test_gradient_root_solving(self, f, df, bounds, roots):
        bound = bounds[0]
        expected = roots[-1]
        tol = 0.01
        for solver in ALL_GRADIENT_BASED_METHODS:
            actual = solver(f, bound[0], tol=tol)
            onp.testing.assert_allclose(expected, actual, atol=tol)


def test_newton_1d():
    expected = math.sqrt(2)
    def f(x):
        return x**2 - 2
    tol = 0.1
    actual = root_solving.newton_1d(f, 1.0, tol=tol)
    onp.testing.assert_allclose(actual, expected, atol=tol)


if __name__ == "__main__":
    jax.config.update("jax_enable_x64", True)
    absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for jaxsnn.funcutils."""

import numpy as np
from absl.testing import absltest, parameterized

from njx.base import funcutils


class TrajectoryTests(parameterized.TestCase):
    @parameterized.named_parameters(
        dict(
            testcase_name="identity",
            trajectory_length=6,
            post_process=funcutils._identity,
        ),
        dict(
            testcase_name="squared_postprocessing",
            trajectory_length=12,
            post_process=lambda x: (x[0] ** 2,),
        ),
    )
    def test_trajectory(self, trajectory_length, post_process):
        def step_fn(x):
            return (x[0] + 1,)

        trajectory_fn = funcutils.trajectory(step_fn, trajectory_length, post_process)

        initial_state = (2 * np.ones(1),)
        expected_frames = []
        frame = initial_state
        for _ in range(trajectory_length):
            frame = step_fn(frame)
            expected_frames.append(post_process(frame))
        expected_output = (np.stack([x[0] for x in expected_frames]),)
        _, actual_output = trajectory_fn(initial_state)
        for expected, actual in zip(expected_output, actual_output):
            np.testing.assert_allclose(expected, actual, atol=1e-9)


if __name__ == "__main__":
    absltest.main()



# Copyright (c) 2022 Heidelberg University. All rights reserved.
#
# Released under Apache 2.0 license as described in the file LICENSE.
# Authors: Christian Pehle

import dataclasses

import jax
import jax.numpy as jnp
import tree_math
from njx.base.types import ArrayLike


def linear_interpolation(f_a, f_b, a, b, x):
    return (x - a) / (b - a) * f_b + (b - x) / (b - a) * f_a


def linear_interpolated_root(f_a, f_b, a, b):
    return (a * f_b - b * f_a) / f_b - f_a


@dataclasses.dataclass
@tree_math.struct
class NewtonState:
    it: int
    x: ArrayLike


def newton_1d(f, x0, tol):
    initial_state = NewtonState(it=0, x=x0)

    def cond(state):
        return jnp.logical_and((jnp.abs(f(state.x)) > tol), state.it < 1e6)

    def body(state):
        it, x = state.it, state.x
        fx, dfx = f(x), jax.grad(f)(x)
        step = fx / (dfx + 0.0001)
        new_state = NewtonState(it + 1, x - step)
        return new_state

    return jax.lax.while_loop(
        cond,
        body,
        initial_state,
    ).x


def newton_nd(f, x0):
    initial_state = (0, x0)

    def cond(state):
        it, x = state
        return it < 10

    def body(state):
        it, x = state
        fx, dfx = f(x), jax.grad(f)(x)
        step = jax.numpy.linalg.solve(dfx, -fx)

        new_state = it + 1, x + step
        return new_state

    return jax.lax.while_loop(
        cond,
        body,
        initial_state,
    )[1]


def bisection(f, x_min, x_max, tol):
    """Bisection root finding method

    Based on the intermediate value theorem, which
    guarantees for a continuous function that there
    is a zero in the interval [x_min, x_max] as long
    as sign(f(x_min)) != sign(f(x_max)).

    NOTE: We do not check the precondition sign(f(x_min)) != sign(f(x_max)) here

    f: function for which we want to find a root in the interval [x_min,x_max].
    x_min: lower bound
    x_max: upper bound
    eps: resolution or tolerance of the root finding method

    """
    initial_state = (0, x_min, x_max)  # (iteration, x)

    def cond(state):
        it, x_min, x_max = state
        return jnp.abs(f(x_min)) > tol  # it > 10

    def body(state):
        it, x_min, x_max = state
        x = (x_min + x_max) / 2

        sfxm = jnp.sign(f(x_min))
        sfx = jnp.sign(f(x))

        x_min = jnp.where(sfx == sfxm, x, x_min)
        x_max = jnp.where(sfx == sfxm, x_max, x)

        new_state = (it + 1, x_min, x_max)
        return new_state

    return jax.lax.while_loop(
        cond,
        body,
        initial_state,
    )[1]


@dataclasses.dataclass
@tree_math.struct
class State:
    a: ArrayLike
    b: ArrayLike
    fa: ArrayLike
    fb: ArrayLike


def illinois_method(f, a, b, eps):
    """Illinois root finding method

    This is a modified version of the secant method. Some version of this is
    used in the SUNDIALS suite.

    Reference: Kathie L. Hiebert and Lawrence F. Shampine, Implicitly Defined Output Points for Solutions of ODEs, Sandia National Laboratory Report SAND80-0180, February 1980.
    TODO: We need to fully check edge cases and introduce a suitable set of benchmark problems.

    f: function for which we want to find a root in the interval [a,b].
    a: lower bound
    b: upper bound
    eps: resolution or tolerance of the root finding method

    """
    fa = f(a)
    fb = f(b)
    (a, fa, b, fb) = jax.lax.cond(
        jnp.abs(fa) > jnp.abs(fb), lambda: (b, fb, a, fa), lambda: (a, fa, b, fb)
    )
    init = State(a=a, b=b, fa=fa, fb=fb)

    def cond(state: State):
        return jnp.abs(state.b - state.a) > eps

    def body_fun(state: State):
        a = state.a
        b = state.b
        fa = state.fa
        fb = state.fb

        c = a - (fa * (b - a)) / (fb - fa)
        fc = f(c)
        b, fb = jax.lax.cond(fa * fc <= 0, lambda: (a, fa), lambda: (b, 0.5 * fb))

        state = State(a=c, b=b, fa=fc, fb=fb)
        return state

    return jax.lax.while_loop(cond, body_fun, init).a



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for time_stepping."""

import jax
import jax.numpy as jnp
import numpy as np
from absl.testing import absltest, parameterized

from njx.base import explicit, funcutils


def harmonic_oscillator(x0, t):
    theta = jnp.arctan(x0[0] / x0[1])
    r = jnp.linalg.norm(x0, ord=2, axis=0)
    return r * jnp.stack([jnp.sin(t + theta), jnp.cos(t + theta)])


ALL_TEST_PROBLEMS = [
    dict(
        testcase_name="_harmonic_oscillator_explicit",
        explicit_terms=lambda x: jnp.stack([x[1], -x[0]]),
        projection=lambda x: x,
        dt=1e-2,
        inner_steps=20,
        outer_steps=5,
        initial_state=np.ones(2),
        closed_form=harmonic_oscillator,
        tolerances=[1e-2, 3e-5, 3e-5, 4e-7],
    ),
]


ALL_TIME_STEPPERS = [
    explicit.forward_euler,
    explicit.midpoint_rk2,
    explicit.heun_rk2,
    explicit.classic_rk4,
]


class TimeSteppingTest(parameterized.TestCase):
    @parameterized.named_parameters(ALL_TEST_PROBLEMS)
    def test_integration(
        self,
        explicit_terms,
        projection,
        dt,
        inner_steps,
        outer_steps,
        initial_state,
        closed_form,
        tolerances,
    ):
        # Compute closed-form solution.
        time = dt * inner_steps * (1 + np.arange(outer_steps))
        expected = jax.vmap(closed_form, in_axes=(None, 0))(initial_state, time)

        # Compute trajectory using time-stepper.
        for atol, time_stepper in zip(tolerances, ALL_TIME_STEPPERS):
            with self.subTest(time_stepper.__name__):
                equation = explicit.ExplicitConstrainedODE(explicit_terms, projection)
                step_fn = time_stepper(equation, dt)
                integrator = funcutils.trajectory(
                    funcutils.repeated(step_fn, inner_steps), outer_steps
                )
                _, actual = integrator(initial_state)
                np.testing.assert_allclose(expected, actual, atol=atol, rtol=0)


if __name__ == "__main__":
    jax.config.update("jax_enable_x64", True)
    absltest.main()



import dataclasses
import jaxsnn.base.implicit as time_stepping
import jax.numpy as np
import tree_math

from jaxsnn.base.types import ArrayLike
from jaxsnn.base.tree_solver import tree_solve, tree_matmul


@tree_math.struct
class NeuronState:
    v: ArrayLike
    I: ArrayLike


@dataclasses.dataclass
class MultiCompartmentNeuronModel(time_stepping.ImplicitExplicitODE):
    d: ArrayLike
    u: ArrayLike
    p: ArrayLike

    # Parameters
    tau_syn_inv: ArrayLike

    def explicit_terms(self, state):
        return NeuronState(v=state.I, I=-self.tau_syn_inv * state.I)

    def implicit_terms(self, state):
        return NeuronState(
            v=tree_matmul(self.d, self.u, self.p, state.v), I=np.zeros_like(state.I)
        )

    def implicit_solve(self, state, step_size):
        return NeuronState(
            v=tree_solve(1 - step_size * self.d, step_size * self.u, self.p, state.v),
            I=state.I,
        )



# Copyright (c) 2022 Heidelberg University. All rights reserved.
#
# Released under Apache 2.0 license as described in the file LICENSE.
# Authors: Christian Pehle
from __future__ import annotations

import dataclasses
from typing import Generic, NamedTuple, Type, TypeVar, Union

import jax
import jax.numpy as jnp
import numpy as np
import tree_math

PyTreeState = TypeVar("PyTreeState")
ArrayLike = Union[jnp.ndarray, np.ndarray, float]
Array = Union[jnp.ndarray, np.ndarray]
JaxArray = Type[jnp.ndarray]


@dataclasses.dataclass
@tree_math.struct
class Spike:
    time: JaxArray
    idx: JaxArray

    @property
    def shape(self):
        return self.time.shape

    def __getitem__(self, key: int) -> Spike:
        return Spike(self.time[key], self.idx[key])


@dataclasses.dataclass
@tree_math.struct
class EventPropSpike:
    time: JaxArray
    idx: JaxArray
    current: JaxArray

    @property
    def shape(self):
        return self.time.shape

    def __getitem__(self, key) -> EventPropSpike:
        return EventPropSpike(self.time[key], self.idx[key], self.current[key])


@dataclasses.dataclass
@tree_math.struct
class InputQueue:
    spikes: EventPropSpike
    head: int = 0

    @property
    def is_empty(self) -> bool:
        return self.head >= self.spikes.time.size

    def peek(self) -> EventPropSpike:
        return self.spikes[self.head]

    def pop(self) -> EventPropSpike:
        spike = self.spikes[self.head]
        self.head += 1
        return spike

    def next_time_or_default(self, default):
        return jax.lax.cond(self.is_empty, lambda: default, lambda: self.peek().time)




State = TypeVar("State")


@dataclasses.dataclass
@tree_math.struct
class StepState(Generic[State]):
    neuron_state: State
    time: float
    input_queue: InputQueue


class WeightInput(NamedTuple):
    input: JaxArray


class WeightRecurrent(NamedTuple):
    input: JaxArray
    recurrent: JaxArray


Weight = WeightInput | WeightRecurrent



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""JAX utility functions."""

import contextlib
from typing import Any, Callable, Sequence

import jax
from jax import tree_util
import jax.numpy as jnp


# Specifying the full signatures of Callable would get somewhat onerous
# pylint: disable=g-bare-generic
# Not accurate for contextmanager
# pylint: disable=g-doc-return-or-yield

# There is currently no good way to indicate a jax "pytree" with arrays at its
# leaves. See https://jax.readthedocs.io/en/latest/jax.tree_util.html for more
# information about PyTrees and https://github.com/google/jax/issues/3340 for
# discussion of this issue.
PyTree = Any


_INITIALIZING = 0


@contextlib.contextmanager
def init_context():
    """Creates a context in which scan() only evaluates f() once.
    This is useful for initializing a neural net with Haiku that involves modules
    that are applied inside scan(). Within init_context(), these modules are only
    called once. This allows us to preserve the pre-omnistaging behavior of JAX,
    e.g., so we can initialize a neural net module pass directly into a scanned
    function.
    """
    global _INITIALIZING
    _INITIALIZING += 1
    try:
        yield
    finally:
        _INITIALIZING -= 1


def _tree_stack(trees: Sequence[PyTree]) -> PyTree:
    if trees:
        return tree_util.tree_map(lambda *xs: jnp.stack(xs), *trees)
    else:
        return trees


def scan(f, init, xs, length=None):
    """A version of jax.lax.scan that supports init_context()."""
    # Note: we use our own version of scan rather than haiku.scan() because
    # haiku.scan() only support use inside haiku modules, but we want to be able
    # to use the same scan function even when not using haiku.
    if _INITIALIZING:
        xs_flat, treedef = tree_util.tree_flatten(xs)
        if length is None:
            (length,) = {x.shape[0] for x in xs_flat}
        x0 = tree_util.tree_unflatten(treedef, [x[0, ...] for x in xs_flat])
        carry, y0 = f(init, x0)
        # Create a dummy-output of the right shape while only calling f() once.
        ys = _tree_stack(length * [y0])
        return carry, ys
    return jax.lax.scan(f, init, xs, length)


def repeated(f: Callable, steps: int) -> Callable:
    """Returns a repeatedly applied version of f()."""

    def f_repeated(x_initial):
        g = lambda x, _: (f(x), None)
        x_final, _ = scan(g, x_initial, xs=None, length=steps)
        return x_final

    return f_repeated


def _identity(x):
    return x


def trajectory(
    step_fn: Callable,
    steps: int,
    post_process: Callable = _identity,
    *,
    start_with_input: bool = False,
) -> Callable:
    """Returns a function that accumulates repeated applications of `step_fn`.
    Args:
      step_fn: function that takes a state and returns state after one time step.
      steps: number of steps to take when generating the trajectory.
      post_process: transformation to be applied to each frame of the trajectory.
      start_with_input: if True, output the trajectory at steps [0, ..., steps-1]
        instead of steps [1, ..., steps].
    Returns:
      A function that takes an initial state and returns a tuple consisting of:
        (1) the final frame of the trajectory _before_ `post_process` is applied.
        (2) trajectory of length `steps` representing time evolution.
    """

    # TODO(shoyer): change the default to start_with_input=True, once we're sure
    # it works for training.
    def step(carry_in, _):
        carry_out = step_fn(carry_in)
        frame = post_process(carry_in if start_with_input else carry_out)
        return carry_out, frame

    def multistep(values):
        return scan(step, values, xs=None, length=steps)

    return multistep


def controlled_trajectory(
    step_fn: Callable,
    xs: PyTree,
    # post_process: Callable = _identity,
    *,
    start_with_input: bool = False,
) -> Callable:
    """Returns a function that accumulates repeated applications of `step_fn`.
    Args:
      step_fn: function that takes a state and returns state after one time step.
      xs: control input to generate the trajectory
      post_process: transformation to be applied to each frame of the trajectory.
      start_with_input: if True, output the trajectory at steps [0, ..., steps-1]
        instead of steps [1, ..., steps].
    Returns:
      A function that takes an initial state and returns a tuple consisting of:
        (1) the final frame of the trajectory _before_ `post_process` is applied.
        (2) trajectory of length `steps` representing time evolution.
    """

    def step(carry_in, x):
        carry_out, out = step_fn(carry_in, x)
        # frame = post_process(carry_in if start_with_input else carry_out)
        return carry_out, out

    def multistep(values):
        return scan(step, values, xs=xs)

    return multistep



# Copyright (c) 2022 Heidelberg University. All rights reserved.
#
# Released under Apache 2.0 license as described in the file LICENSE.
# Authors: Christian Pehle

import jax.numpy as jnp
import jax.lax as lax
from functools import partial
from njx.base.types import JaxArray
import tree_math


def tree_to_matrix(d, u, p):
    """
    Auxiliary function, which turns a 'tree' matrix into
    a dense matrix. This is mostly convenient for testing.
    """
    N = d.shape[0]
    a = jnp.diag(d, 0)

    def body_fun(i, val):
        val = val.at[p[i], i].set(u[i - 1])
        val = val.at[i, p[i]].set(u[i - 1])
        return val

    # for i in range(1, N):
    #     a = a.at[p[i], i].set(u[i - 1])
    #     a = a.at[i, p[i]].set(u[i - 1])

    return lax.fori_loop(1, N, body_fun, a)



@tree_math.struct
class TreeMatrix:
    d: JaxArray
    u: JaxArray
    p: JaxArray


@tree_math.struct
class TreeProblem:
    t: TreeMatrix
    b: JaxArray


@tree_math.struct
class TreeMatmulProblem:
    t: TreeMatrix
    b: JaxArray
    r: JaxArray


def tree_matmul(d, u, p, b):
    """
    Multiply a 'tree' matrix with a vector.

    The only non-zero entries are of the form

    a[p[i], i], a[i,i], a[i, p[i]]

    however in each row there can be many entries

    m[i] = a[i, p[i]] * b[p[i]]
    n[p[i]] = a[p[i], i] * b[i]
    diag = d * b

    """
    def body_fun(i, val):
        val.r = val.r.at[i].add(val.t.u[i - 1] * val.b[val.t.p[i]])
        val.r = val.r.at[val.t.p[i]].add(val.t.u[i - 1] * val.b[i])
        return val

    N = d.shape[0]

    init_val = TreeMatmulProblem(t=TreeMatrix(u=u, d=d, p=p), b=b, r=d * b)

    return lax.fori_loop(1, N, body_fun, init_val).r


def hines_solver(d, u, p, b):
    """ """
    N = d.shape[0]

    def reverse_body(j, val: TreeProblem):
        i = N - j - 1
        f = val.t.u[val.t.p[i]] / val.t.d[i]
        val.t.d = val.t.d.at[val.t.p[i]].set(
            val.t.d[val.t.p[i]] - f * val.t.u[val.t.p[i]]
        )
        val.b = val.b.at[val.t.p[i]].set(val.b[val.t.p[i]] - f * val.b[i])

        return val

    def forward_body(i, val: TreeProblem):
        val.b = val.b.at[i].set(
            (val.b[i] - val.t.u[val.t.p[i]] * val.b[val.t.p[i]]) / val.t.d[i]
        )
        return val

    problem = TreeProblem(t=TreeMatrix(d=d, u=u, p=p), b=b)
    res = lax.fori_loop(0, N - 1, reverse_body, init_val=problem)
    res.b = res.b.at[0].set(res.b[0] / res.t.d[0])
    return lax.fori_loop(1, N, forward_body, init_val=res).b


def tree_solve(d, u, p, b):
    """
    A solver for 'tree' matrices, which is compatible with the jax tracer.
    """
    solver = partial(
        lax.custom_linear_solve,
        lambda x: tree_matmul(d, u, p, x),
        solve=lambda _, x: hines_solver(d, u, p, x),
        symmetric=True,
    )

    return solver(b)
